-- BITONIC SORT
-- Emil Falk
-- emil.falk.1988@gmail.com
--
-- Most code taken from Simon Marlows paper:
-- http://community.haskell.org/~simonmar/papers/threadscope.pdf
--
-- MODIFICATIONS:
-- First off, instead of recomputing and traversing lists all the time
-- to find out it's length compute the length in the beginning and pass
-- that around. Saves some computation time. The original implementation
-- is a veeeery naive parallel implementation where there are millions
-- of sparks. To change this I introduce thresholds for both places where
-- parallelism occurs.
--
-- Bitonic sort can only sort 2^n where n>1 elements each time.
--
-- NOTE: To make it work with arbitrary sized lists I take the easiest
--       approach. When computing the length, find the maximum element
--       and pad the input list with the maximum element until the list
--       has a power of two.
--
-- NOTE: I've not fine tuned the thresholds and I do not think this is in
--       anyway optimal.

module Bitonic
       (sort)
       where

import Control.Parallel
import Control.Parallel.Strategies
import Control.DeepSeq

-- Thresholds for when we revert to sequential evaluation
thPar2, thEvens :: Int
thPar2  = 512
thEvens = 512

-- Function composition reversed
f >- g = g . f

-- Parallel evaluation of a pair of lists
-- When the input-lists are less than or equal to th_pars we
-- revert to sequential evaluation
par2 :: NFData a => Int -> -- Size of each input list
       ([a] -> [a]) ->    -- Circuit 1
       ([a] -> [a]) ->    -- Circuit 2
       ([a],[a]) ->      -- Input
       ([a],[a])        -- Output
par2 m circuit1 circuit2 (input1, input2)
  | m <= thPar2 = (output1,output2)
  | otherwise   = (output1,output2) `using` parTuple2 rdeepseq rdeepseq
  where
    output1 = circuit1 input1
    output2 = circuit2 input2

-- Halve a list, do parallel evaluation of r on the two sub-lists
-- and concatenate the results
two :: NFData a => ([a] -> [a]) -> Int -> [a] -> [a]
two r m = halve m >- par2 m r r >- unhalve

-- Uninterleave a list, do a parallel pair evaluation on
-- both halves and interleave the results
ilv :: NFData a => ([a] -> [a]) -> Int -> [a] -> [a]
ilv r m = unriffle >- two r m >- riffle m

-- Halve a list
halve :: Int -> [a] -> ([a],[a])
halve = splitAt

-- Unhalve a list
unhalve :: ([a],[a]) -> [a]
unhalve = uncurry (++)

-- Halve a list, then interleave the two list into one
riffle :: Int -> [a] -> [a]
riffle m = halve m >- uncurry zip >- unpair
  where unpair [] = []
        unpair ((x,y):xs) = x : y : unpair xs

-- Inverse of riffle, uninterleave the list
unriffle :: [a] -> [a]
unriffle = pair >- unzip >- unhalve
  where pair [] = []
        pair (x:y:xs) = (x,y):pair xs

-- A parallel butterfly network used for the merging
butterfly :: NFData a => ([a] -> [a]) -> Int -> [a] -> [a]
butterfly circuit _ l@[x,y] = circuit l
butterfly circuit n input
  = (ilv (butterfly circuit m) m >- evens circuit n) input
    where m = n `div` 2

-- Chop the input list into sub-lists of size 2. Map f on each sublist
-- (compare) and concatenate the results
-- NOTE: Maybe let the number of chunks depend on some threshold?
evens :: NFData a => ([a] -> [a]) -> Int -> [a] -> [a]
evens f m | m < thEvens = chop >- map f >- concat
          | otherwise =
            chop >- (withStrategy (parListChunk 4 rdeepseq) . map f) >- concat

-- Chop a list into sublists of size 2
chop :: [a] -> [[a]]
chop [] = []
chop (x:y:xs) = [x,y]:chop xs

-- Bitonic sort
-- Takes a compare-function, the size of the input list and an input
-- list and the input lists. The size has to be a power of 2.
-- Returns the input lists sorted.
sortB :: NFData a => ([a] -> [a]) -> Int -> [a] -> [a]
sortB cmp _ l@[x,y] = cmp l
sortB cmp n input
  = (two (sortB cmp m) m >- sndList m reverse >- butterfly cmp n) input
    where m = n `div` 2

-- Bitonic sort
-- sort takes an input list to sort. If this list does not have a length
-- of power of 2 then pad it with the maximum element in the list, sort the
-- padded list and drop the padding before returning the sorted list.
-- NOTE: Wierd "bug", cannot call sortB cmp n xs' directly, why?
--       Tried to use seq to make xs' strict but to no avail. This
--       seems to work atleast.
sort :: (NFData a, Ord a) => [a] -> [a]
sort xs = if pow-n == 0
          then xs `seq` sortB cmp n xs
          else take n (sort xs')
  where (n,m) = lengthMax xs
        -- Naively compute the nearest power of two larger than n
        pow :: Int
        pow   = head . dropWhile (<n) . map (2^) $ [1..]
        xs' = xs ++ replicate (pow-n) m

-- Compare function
cmp :: Ord a => [a] -> [a]
cmp [x,y] = if x < y then [x,y] else [y,x]

-- Split a list into two and apply a function to the upper half then
-- concatenate the results
sndList :: Int -> ([a] -> [a]) -> [a] -> [a]
sndList m f xs = f l ++ r
  where (l,r) = splitAt m xs

-- Compute the length and find the maximum element in one sweep.
lengthMax :: Ord a => [a] -> (Int,a)
lengthMax (x:xs) = foldl
                   (\(l,m) e -> if e > m then (l+1,e) else (l+1,m))
                   (1,x) xs