module MergeSort where

import Control.Parallel
import Control.DeepSeq
import Data.List
import Data.Maybe
import Debug.Trace (trace)


{- Non parallel merge sort: used as a baseline and as a fallback in the parallel one. -}
mergesort :: Ord a => [a] -> [a]
mergesort [] = []
mergesort [a] = [a]
mergesort l = merge (mergesort l1) (mergesort l2)
  where (l1,l2) = splitAt (length l `div` 2) l

merge [] bs = bs
merge as [] = as
merge (a:as) bs@(b:_) | a < b = a:merge as bs
merge as (b:bs) = b:merge as bs


{- Parallel version of merge sort. It is the exact same algorithm as above exept that
   it evaluates the two half of the list in parallel.
   After a certain depth, it reverts on the sequential version to avoid creating too much   threads. -}
pmergesort :: Ord a => Int -> [a] -> [a]
pmergesort depth xs = pmergesort' depth (length xs) xs

pmergesort' :: Ord a => Int -> Int -> [a] -> [a]
pmergesort' _ _ [] = []
pmergesort' _ _ [a] = [a]
pmergesort' 0 len xs = mergesort xs
pmergesort' d len xs = 
  par (length premier) $
  pseq (length second) $
  pmerge d premier second
  where n = len `div` 2
        (xs1,xs2) = splitAt n xs
        premier =  pmergesort' (d-1) n xs1
        second = pmergesort' (d-1) (len-n) xs2

{- A parallel merge that split the list it two and merge them independently.  
   A concatenation of the two merged sublists is returned. -}
pmerge :: Ord a => Int -> [a] -> [a] -> [a]
pmerge 0 as bs = merge as bs
pmerge d as bs = 
  par (length part1) $
  pseq (length part2) $
  part1 ++ part2
  where m = length as 
        (as1,as2) = splitAt (m `div` 2) as
        a = head as2
        (bs1,bs2) = break (>a) bs
        part1 = pmerge (d-1) as1 bs1
        part2 = pmerge (d-1) as2 bs2

{- Failed tentative: One of the main problem in the above algorithm is space complexity.
   This algorithm on the other hand is much more efficient in term of space,
   but it's also much harder to parallelize… Augmenting the inital size of the heap
   gave better results :-D -}
mergesort2 :: Ord a => [a] -> [a]
mergesort2 = mergesort2' 0 . map (:[])

mergesort2' :: Ord a => Int -> [[a]] -> [a]
mergesort2' _ [] = []
mergesort2' _ [a] = a
mergesort2' d l | d > 10 = mergesort2' (d+1) $ merge2 d l
mergesort2' d l = 
  par (length $ merge2 d l) $
  mergesort2' (d+1) $ merge2 d l

merge2 :: Ord a => Int -> [[a]] -> [[a]]
merge2 _ [] = []
merge2 _ [l] = [l]
merge2 d (l1:l2:ls) | d < 10000000 = l12:cont
  where l12 = merge l1 l2
        cont = merge2 d ls
merge2 d (l1:l2:ls) = 
  par (length l12) $ 
  pseq (length cont) $
  l12:cont
  where l12 = merge l1 l2
        cont = merge2 d ls
