{-# LANGUAGE BangPatterns #-}
{-# OPTIONS -O2 -fspec-constr-count=20 #-}


import System.Random
import qualified Data.List as DL
import Control.DeepSeq
import Control.Parallel
import Control.Parallel.Strategies
import Criterion.Main
import Test.QuickCheck
import Data.Int

randomIntegers = take 200000 (randoms (mkStdGen 211570155)) :: [Integer]

-- | quicksort based on Ken Friis Larsen's algorithm from
-- http://ken.friislarsen.net/blog/2011/09/28/quality-assessment-haskell-program/
--
qsortk :: Ord a => [a] -> [a]
qsortk xs = go xs
  where
    go []     = []
    go [x]    = [x]
    go [x,y]  = if x > y then [y,x] else [x,y]
    go (p:xs) = go lesser ++ equal ++ go greater
      where
        (lesser, equal, greater) = DL.foldl' part ([], [p], []) xs
        part (!l, !e, !g) x =
          case compare x p of
            LT -> (x:l,   e,   g)
            GT -> (  l,   e, x:g)
            EQ -> (  l, x:e,   g)

qsortkp :: [Integer] -> [Integer]
qsortkp xs = go xs
  where
    go [] = []
    go [x] = [x]
    go [x,y] = if x > y then [y,x] else [x,y]
    go (p:xs) = rnf g `par` e `par` l `pseq` (l ++ e ++ g)
      where
        l = go lesser
        e = equal
        g = go greater
        (lesser, equal, greater) = DL.foldl' part ([], [p], []) xs
        part (!l, !e, !g) x =
          case compare x p of
            LT -> (x:l,   e,   g)
            GT -> (  l,   e, x:g)
            EQ -> (  l, x:e,   g)

qsortkpd :: Integer -> [Integer] -> [Integer]
qsortkpd limit xs = go limit xs
  where
    go _ []     = []
    go _ [x]    = [x]
    go _ [x,y]  = if x > y then [y,x] else [x,y]
    go 0 xs     = qsort3 xs
    go d (p:xs) = rnf g `par` e `par` l `pseq` (l ++ e ++ g)
      where
        l = go (d-1) lesser
        e = equal
        g = go (d-1) greater
        (lesser, equal, greater) = DL.foldl' part ([], [p], []) xs
        part (!l, !e, !g) x =
          case compare x p of
            LT -> (x:l,   e,   g)
            GT -> (  l,   e, x:g)
            EQ -> (  l, x:e,   g)

qsortkpl :: Int64 -> [Integer] -> [Integer]
qsortkpl limit xs = go (maxBound :: Int64) xs
  where
    go _   []     = []
    go _   [x]    = [x]
    go _   [x,y]  = if x > y then [y,x] else [x,y]
    go len xs | len < limit = qsort3 xs
    go len (p:xs) = rnf g `par` e `par` l `pseq` (l ++ e ++ g)
      where
        l = go ll lesser
        e = equal
        g = go gl greater
        (ll, lesser, equal, gl, greater) = DL.foldl' part (0, [], [p], 0, []) xs
        part (!ll, !l, !e, !gl, !g) x =
          case compare x p of
            LT -> (1+ll, x:l,   e,   gl,   g)
            GT -> (  ll,   l,   e, 1+gl, x:g)
            EQ -> (  ll,   l, x:e,   gl,   g)

-- | QuickSort with an accumulator, an implementation by Will Ness from
-- http://stackoverflow.com/questions/7641936/is-it-possible-to-do-quicksort-of-a-list-with-only-one-passing
--
qsort3 xs = qcat xs []
  where
    qcat (x:xs) zs = part x xs zs [] [] []
    qcat []     zs = zs
    qapp (x:xs) zs = x:qapp xs zs
    qapp []     zs = zs
    part x []     zs !a !b !c = qcat a $ qapp (x : b) $ qcat c zs
    part x (y:ys) zs !a !b !c =
        case compare y x of
                  LT -> part x ys zs (y:a) b c
                  EQ -> part x ys zs a (y:b) c
                  GT -> part x ys zs a b (y:c)

main = defaultMain
    [ bgroup "haskell"
        [ bench "DL.sort" (nf DL.sort randomIntegers)
        ]
    , bgroup "quicksort"
        [ -- bench "qsortk" (nf qsortk randomIntegers)
--        , bench "qsortkp" (nf qsortkp randomIntegers)
--        , bench "qsortkpd (1)" (nf (qsortkpd 1) randomIntegers)
--        , bench "qsortkpd (2)" (nf (qsortkpd 2) randomIntegers)
         bench "qsortkpd (3)" (nf (qsortkpd 3) randomIntegers)
--        , bench "qsortkpd (5)" (nf (qsortkpd 5) randomIntegers)
--        , bench "qsortkpl (256)" (nf (qsortkpl 256) randomIntegers)
--        , bench "qsortkpl (512)" (nf (qsortkpl 512) randomIntegers)
--        , bench "qsortkpl (1024)" (nf (qsortkpl 1024) randomIntegers)
--        , bench "qsortkpl (2048)" (nf (qsortkpl 2048) randomIntegers)
--        , bench "qsortkpl (4096)" (nf (qsortkpl 4096) randomIntegers)
--        , bench "qsortkpl (8192)" (nf (qsortkpl 8192) randomIntegers)
--        , bench "qsort3" (nf qsort3 randomIntegers)
        ]

    ]

sorted :: Ord a => [a] -> Bool
sorted (x1:x2:xs) = x1 <= x2 && (sorted $ x2:xs)
sorted _          = True

prop_sorted :: Ord a => ([a] -> [a]) -> [a] -> Bool
prop_sorted sorter xs = sorted $ sorter xs

tests = do
  quickCheck $ prop_sorted (qsortk          :: [Integer] -> [Integer])
  quickCheck $ prop_sorted (qsortkp         :: [Integer] -> [Integer])
  quickCheck $ prop_sorted (qsortkpd 4      :: [Integer] -> [Integer])
  quickCheck $ prop_sorted (qsortkpl 2048   :: [Integer] -> [Integer])
  quickCheck $ prop_sorted (qsort3          :: [Integer] -> [Integer])


