{-# LANGUAGE ScopedTypeVariables #-}

module Main where

import Control.DeepSeq
import Control.Monad
import Control.Parallel
import Control.Parallel.Strategies
import Criterion.Main
import Data.Bits
import Data.Function
import Data.List hiding (insert)
import System.Random
import Test.QuickCheck
import qualified Data.DList as DL
import qualified Data.IntMap as IM

main :: IO ()
main = do
  runTests
  defaultMain $
--    [ bench "msort"  (nf msort randomInts) ] ++
  --  [ bench "bsort"  (nf bsort randomInts) ] ++
    [ bench "bsort'" (nf bsort'IntegerCheating  randomInts) ]
-- ++
--    [ bench "imsort" (nf (imsort 16) randomInts) ]

-- Naive merge sort
msort :: [Integer] -> [Integer]
msort xs = go (6 :: Int) (len, xs)
  where
    len = length xs
    go 0 (_, xs)  = sort xs
    go _ (_, [])  = []
    go _ (_, [x]) = [x]
    go d (n, xs)  = runEval $ do
      let newD = d-1
      let (as,bs) = split (n,xs)
      as' <- rpar `dot` rdeepseq $ (go newD as)
      return $ merge as' (go newD bs)
    split (n,xs) = ((halfN, as), (if odd n then halfN + 1 else halfN, bs))
      where
        (as,bs) = splitAt halfN xs
        halfN = div n 2

merge as@(a:as') bs@(b:bs')
  | a > b     = b:merge as  bs'
  | otherwise = a:merge as' bs
merge [] bs   = bs
merge as []   = as


-- IntMap sort
{-
imsort :: Int -> [Integer] -> [Integer]
imsort _ [] = []
imsort _ [x] = [x]
imsort n xs = map toInteger res
  where
    xs'   = map fromInteger xs
    csize = max 1 (length xs `div` n)
    lists = chunks csize xs'
    res   = imToList $ IM.unionsWith (+)
              (parMap rdeepseq (IM.fromListWith (+) . (`zip` repeat 1)) lists)

    chunks :: Int -> [a] -> [[a]]
    chunks n = map (take n) . takeWhile (not . null) . iterate (drop n)

    imToList :: IM.IntMap Int -> [Integer]

    imToList im = DL.toList (IM.foldrWithKey'
                              (\i n -> (DL.replicate n (toInteger i) `DL.append`))
                              DL.empty im)
-}

-- radix sort on the top `n' bits
bsort :: [Integer] -> [Integer]
bsort list = map toInteger $
             pseq
               (lists `using` parListChunk (2^(n-2)) rdeepseq)
               (concat (buckets `using` parListChunk 4 rdeepseq))
  where
    n = 4
    list' = map fromInteger list :: [Int]
    testBits = [63,62..63-n+1]
    bools    = sortBy (flip compare `on` head) $ replicateM n [False,True]
    bitBools = map (zip testBits) bools
    lists    = [ [i|i <- list', all (\(n,b) -> testBit i n == b) nbs]
               | nbs <- bitBools ]
    buckets = map sort lists

bsort'IntegerCheating :: [Integer] -> [Integer]
bsort'IntegerCheating xs = map toInteger
                        $ (bsort' :: [Int] -> [Int])
                        $ map fromInteger
                        $ xs

-- Variation on radix sort.
bsort' :: forall a . (NFData a,Bits a,Ord a) => [a] -> [a]
bsort' xs = DL.toList (bucket t xs)
  where
    t :: Int
    t = 2

    bits :: Int
    bits = bitSize (undefined :: a)

    swap (a,b) = (b,a) -- uncurry (flip (,))

    bucket :: Int -> [a] -> DL.DList a
    bucket 0 xs = let s = sort xs in rnf s `pseq` DL.fromList s
    bucket d xs = let (u,l) = (if d == t then swap else id)
                              (partition (`testBit` (bits - t - 1 + d)) xs)
                      us = bucket (d-1) u
                      ls = bucket (d-1) l
                  in  us `par` ls `pseq` (ls `DL.append` us)


----------------------------------------------------------------------
-- Misc

randomInts :: [Integer]
randomInts = take 200000 (randoms (mkStdGen 211570155))

prop_bsort, prop_msort :: [Integer] -> Bool
prop_bsort as = bsort as == sort as
prop_msort as = msort as == sort as

--prop_imsort :: [Integer] -> Property
--prop_imsort xs = forAll (choose (2,10)) $ \n -> imsort n xs == sort xs

prop_bsort' :: [Integer] -> Bool
prop_bsort' xs = bsort'IntegerCheating xs == sort xs

runTests :: IO ()
runTests = do
  quickCheck prop_bsort
  quickCheck prop_msort
--  quickCheck prop_imsort
  quickCheck prop_bsort'
