{-# LANGUAGE ScopedTypeVariables #-}

module Main where

import Control.DeepSeq
import Control.Monad
import Control.Parallel
import Control.Parallel.Strategies
import Criterion.Config
import Criterion.Main
import Data.Bits
import Data.Char (isDigit)
import Data.Function
import Data.List hiding (insert)
import System.Environment
import System.Exit
import System.Random
import Test.QuickCheck
import qualified Tree  as T
import qualified Data.DList as DL
import qualified Data.IntMap as IM
import qualified Data.IntSet as IS

main = do
  args <- getArgs
  case args of
    [] -> runTests >> selectBench defaultConfig defaultLen
    _  -> do
      (conf,_) <- parseArgs defaultConfig defaultOptions args
      runTests >> selectBench conf defaultLen
    where
    defaultLen = 200000

selectBench conf len = do
  putStrLn $ "\nCurrent length of 'randomInts' is: " ++ show len
  n <- getOpt
  defaultMainWith conf (return ()) ([tests!!n])
  selectBench conf len
  where
    rands = take len randomInts
    choiceLine =
      "\nChoose a test to run:\n\n" ++
      concatMap (\(n,t) -> show n ++ ") " ++ show t ++ "\n") (zip [0..] tests) ++
      "l) Change length of 'randomInts'\n" ++
      "q) Quit\n"
    getOpt = do
      putStrLn choiceLine
      l <- getLine
      case l of
        "q" -> exitSuccess
        "l" -> getLen >>= selectBench conf
        _ | not (null l) && all isDigit l -> return (read l)
        _   -> getOpt
    getLen = do
      putStrLn "New length: "
      l <- getLine
      if not (null l) && all isDigit l then return (read l) else getLen
    tests = [ bench "psort" (nf psort rands) ] ++
            [ bench "bsort" (nf bsort rands) ] ++
            [ bench ("cmsort " ++ show n) (nf (cmsort n) rands)
            | n <- [4,8..16] ] ++
            [ bench "msort" (nf msort rands) ] ++
            [ bench ("isort " ++ show n) (nf (isort n) rands)
            | n <- [2..8] ] ++
            [ bench ("bucketSort")  (nf bucketSortIntegerCheating  rands) ] ++
            [ bench ("bucketSort' " ++ show t) (nf (bucketSortIntegerCheating' t) rands)
            | t <- [2..8] ] ++
            [ bench "sort"    (nf sort rands) ] ++
            [ bench "sortInt" (nf sortInt rands) ] ++
            [ bench "rsort"   (nf rsort   (map fromInteger rands :: [Int])) ] ++
            [ bench ("issort" ++ show n)   (nf (issort n) rands)
            | n <- [4,8..16] ] ++
            [ bench ("imsort" ++ show n)   (nf (imsort n) rands)
            | n <- [4,8..16] ]


len = 200000
randomInts = take len (randoms (mkStdGen 211570155)) :: [Integer]

issort :: Int -> [Integer] -> [Integer]
issort _ [] = []
issort _ [x] = [x]
issort n xs = map toInteger res
  where
    xs'   = map fromInteger xs
    csize = max 1 (length xs `div` n)
    lists = chunks csize xs'
    res   = IS.toList $ IS.unions (parMap rdeepseq IS.fromList lists)

----------------------------------------------------------------------
-- Parallel sort functions

psort []     = []
psort (x:xs) = par rest $
                    psort [ y | y <- xs, y<x ]
                 ++ [x]
                 ++ rest
  where rest = psort [ y | y <- xs, y >= x ]

psort2 0 xs     = sort xs
psort2 n []     = []
psort2 n (x:xs) = par rest $
                       psort2 (n-1) [ y | y <- xs, y<x ]
                    ++ [x]
                    ++ rest
  where rest = psort2 (n-1) [ y | y <- xs, y >= x ]

-- radix sort on the top `n' bits
bsort list = map toInteger $
             pseq
               (lists `using` parListChunk (2^(n-2)) rdeepseq)
               (concat (buckets `using` parListChunk 4 rdeepseq))
  where
    n = 4
    list' = map fromInteger list :: [Int]
    testBits = [63,62..63-n+1]
    bools    = sortBy (flip compare `on` head) $ replicateM n [False,True]
    bitBools = map (zip testBits) bools
    lists    = [ [i|i <- list', all (\(n,b) -> testBit i n == b) nbs]
               | nbs <- bitBools ]
    buckets = map sort lists

msort xs = go 6 (len, xs)
  where
    len = length xs
    go 0 (_, xs)  = sort xs
    go _ (_, [])  = []
    go _ (_, [x]) = [x]
    go d (n, xs)  = runEval $ do
      let newD = d-1
      let (as,bs) = split (n,xs)
      as' <- rpar `dot` rdeepseq $ (go newD as)
      bs' <- rdeepseq (go newD bs)
      return $ merge as' bs'
    split (n,xs) = ((halfN, as), (if odd n then halfN + 1 else halfN, bs))
      where
        (as,bs) = splitAt halfN xs
        halfN = div n 2

cmsort _ []  = []
cmsort _ [x] = [x]
cmsort n xs  = map toInteger $ mergeAll (sorted `using` parList rdeepseq)
  where
    len    = length xs
    csize  = max 1 (len `div` n)
    lists  = chunks csize (map fromInteger xs :: [Int])
    sorted = map sort lists

isort _ []  = []
isort _ [x] = [x]
isort n xs  = map toInteger $ DL.toList $ runEval $ do
  let xs' = map fromInteger xs :: [Int]
  lists   <- rdeepseq $ chunks (max 4 (length xs' `div` n)) xs'
  let buckets = parMap rdeepseq (foldr (\a -> IM.insertWith (+) a 1) IM.empty) lists
  kvs     <- rdeepseq $ IM.toList $ foldr1 (IM.unionWith (+)) buckets
  return $ foldr (\(i,n) l -> DL.replicate n i `DL.append` l) DL.empty kvs

imsort :: Int -> [Integer] -> [Integer]
imsort _ [] = []
imsort _ [x] = [x]
imsort n xs = map toInteger res
  where
    xs'   = map fromInteger xs
    csize = max 1 (length xs `div` n)
    lists = chunks csize xs'
    res   = imToList $ IM.unionsWith (+) (parMap rdeepseq (IM.fromListWith (+) . (`zip` repeat 1)) lists)

imToList :: IM.IntMap Int -> [Integer]
imToList = DL.toList . IM.foldrWithKey' (\i n l -> DL.replicate n (toInteger i) `DL.append` l) DL.empty

immergeAll :: [IM.IntMap Int] -> IM.IntMap Int
immergeAll [x] = x
immergeAll xs  = immergeAll (immergePairs xs)

immergePairs (a:b:xs) = immerge a b : immergePairs xs
immergePairs xs       = xs

immerge = IM.unionWith (+)


bucketSortIntegerCheating :: [Integer] -> [Integer]
bucketSortIntegerCheating xs = map toInteger
                             $ (bucketSort :: [Int] -> [Int])
                             $ map fromInteger
                             $ xs

bucketSortIntegerCheating' :: Int -> [Integer] -> [Integer]
bucketSortIntegerCheating' d = map toInteger
                             . (bucketSort' d :: [Int] -> [Int])
                             . map fromInteger

bucketSort :: forall a . (NFData a,Bits a,Ord a) => [a] -> [a]
bucketSort xs = DL.toList (bucket t xs)
  where
    t :: Int
    t = 2

    bits :: Int
    bits = bitSize (undefined :: a)

    bucket :: Int -> [a] -> DL.DList a
    bucket 0 xs = let s = sort xs in rnf s `pseq` DL.fromList s
    bucket d xs = let (u,l) = (if d == t then swap else id)
                              (partition (`testBit` (bits - t - 1 + d)) xs)
                      us = bucket (d-1) u
                      ls = bucket (d-1) l
                  in  us `par` ls `pseq` (ls `DL.append` us)

bucketSort' :: forall a . (NFData a,Bits a,Ord a) => Int -> [a] -> [a]
bucketSort' t xs = buckets `pseq` concat (parMap rdeepseq sort buckets)
  where
    bits :: Int
    bits = bitSize (undefined :: a)

    buckets :: [[a]]
    buckets = bucket t xs `using` parListChunk (2^(max (t-2) 2)) rdeepseq

    bucket :: Int -> [a] -> [[a]]
    bucket 0 xs = [xs]
    bucket d xs = let (u,l) = (if d == t then swap else id)
                              (partition (`testBit` (bits - t - 1 + d)) xs)
                  in  bucket (d-1) l ++ bucket (d-1) u

msortInteger :: [Integer] -> [Integer]
msortInteger xs = map toInteger (msortInt (map fromInteger xs))

msortInt :: [Int] -> [Int]
msortInt = msort2

msort2 :: (NFData a,Ord a) => [a] -> [a]
msort2 xs = mergeAll (sequences xs)
  where
    sequences (a:b:xs)
      | a `compare` b == GT = descending b [a]  xs
      | otherwise           = ascending  b (a:) xs
    sequences xs = [xs]

    descending a as (b:bs)
      | a `compare` b == GT = descending b (a:as) bs
    descending a as bs      = (a:as): sequences bs

    ascending a as (b:bs)
      | a `compare` b /= GT = ascending b (\ys -> as (a:ys)) bs
    ascending a as bs       = as [a]: sequences bs

    mergeAll [x] = x
    mergeAll xs  = runEval $ do p  <- rpar (mergePairs xs)
                                rseq (mergeAll p)

    mergePairs (a:b:xs) = runEval $ do m  <- rpar (merge a b)
                                       ps <- rseq (mergePairs xs)
                                       return (m : ps)
    mergePairs xs       = xs

    merge as@(a:as') bs@(b:bs')
      | a `compare` b == GT = b:merge as  bs'
      | otherwise           = a:merge as' bs
    merge [] bs             = bs
    merge as []             = as



----------------------------------------------------------------------
-- Non parallel sorting functions

qsort []     = []
qsort (x:xs) =    qsort [ y | y <- xs, y <  x ]
               ++ [x]
               ++ qsort [ y | y <- xs, y >= x ]

mergesort' :: Ord a => [[a]] -> [a]
mergesort' [] = []
mergesort' [xs] = xs
mergesort' xss = mergesort' (merge_pairs xss)
  where
    merge_pairs :: Ord a => [[a]] -> [[a]]
    merge_pairs [] = []
    merge_pairs [xs] = [xs]
    merge_pairs (xs:ys:xss) = merge xs ys : merge_pairs xss

    merge :: Ord a => [a] -> [a] -> [a]
    merge [] ys = ys
    merge xs [] = xs
    merge (x:xs) (y:ys)
     = case x `compare` y of
            GT -> y : merge (x:xs) ys
            _  -> x : merge xs     (y:ys)

sortInt :: [Integer] -> [Integer]
sortInt = map toInteger
        . (sort :: [Int] -> [Int])
        . map fromInteger

rsort :: Ord a => [a] -> [a]
rsort xs = mergesort' (risers xs)
  where
    risers :: Ord a => [a] -> [[a]]
    risers [] = []
    risers [x] = [[x]]
    risers (x:y:etc) = if x <= y then (x:s):ss else [x]:(s:ss)
      where (s:ss) = risers (y:etc)


----------------------------------------------------------------------
-- misc functions

chunks :: Int -> [a] -> [[a]]
chunks n = map (take n) . takeWhile (not . null) . iterate (drop n)

mergeAll [x] = x
mergeAll xs  = mergeAll (mergePairs xs)

mergePairs (a:b:xs) = runEval $ do
  x  <- rpar $ merge a b
  return $ x : mergePairs xs
mergePairs xs       = xs

merge as@(a:as') bs@(b:bs')
  | a > b     = b:merge as  bs'
  | otherwise = a:merge as' bs
merge [] bs   = bs
merge as []   = as

{-# INLINE swap #-}
swap (x,y) = (y,x)


----------------------------------------------------------------------
-- Tests

prop_bsort, prop_msort :: [Integer] -> Bool
prop_bsort as = bsort as == sort as
prop_msort as = msort as == sort as

prop_cmsort :: [Integer] -> Property
prop_cmsort xs = forAll (choose (1,12)) $ \t -> cmsort t xs == sort xs

prop_chunks :: Int -> [Integer] -> Property
prop_chunks n xs = n > 0 ==> xs == concat (chunks n xs)

prop_isort :: [Integer] -> Property
prop_isort xs = forAll (choose (2,10)) $ \n -> isort n xs == sort xs

prop_imsort :: [Integer] -> Property
prop_imsort xs = forAll (choose (2,10)) $ \n -> imsort n xs == sort xs

prop_issort :: [Integer] -> Property
prop_issort xs = forAll (choose (2,10)) $ \n -> issort n xs == sort xs

prop_bucketSort :: [Int] -> Bool
prop_bucketSort xs = bucketSort xs == sort xs

prop_bucketSort' :: [Int] -> Property
prop_bucketSort' xs = forAll (choose (0,14)) $ \t -> bucketSort' t xs == sort xs

prop_rsort :: [Int] -> Bool
prop_rsort xs = rsort xs == sort xs

prop_msort2 :: [Int] -> Bool
prop_msort2 xs = msort xs == sort xs

runTests = do
  quickCheck prop_bsort
  quickCheck prop_msort
  quickCheck prop_chunks
  quickCheck prop_cmsort
  quickCheck T.prop_fromTo
  quickCheck prop_isort
  quickCheck prop_issort
  quickCheck prop_imsort
  quickCheck prop_bucketSort
  quickCheck prop_bucketSort'
  quickCheck prop_rsort
  quickCheck prop_msort2
