{- 
 Karl Schmidt -- karsch@student

 Sorts the dataset by bucketing the input into an immutable array, sorting each 
 bucket with Data.List.sort and finally concatenating the buckets back into a 
 list. Obviously this doesn't generalize to any data type implementing Ord, 
 although I could probably have coaxed it to anything implementing Ix and 
 Integral if I wanted. Sticking to Integer saved me from trying to solve 
 overflow issues, however.
 
 The concatenation function handles parallel evaluation, using Strategies. 
 Using Strategies badly, even. I should probably have stuck to par and pseq, 
 but hopefully this still passes muster.
  
 I'm a bit surprised at how bad the running time is, I expected bucketsort to 
 be an order of magnitude or so faster than a comparison sort for the given 
 dataset. Looking at threadscope most of the running time is spent on the 
 very sequential dividing of things into buckets and 50% of that is 
 GC, so perhaps I should be less surprised...
-}

--module Bucketsort where

import Control.Parallel.Strategies
-- for some reason Strategies doesn't export NFData in windows, so...
import Control.DeepSeq 
import GHC.Conc

import Criterion.Main
import System.Random

import Data.Array
import Data.List

import Test.QuickCheck

{-
 So, the definition in System.Random goes

 instance Random Integer where
   randomR ival g = randomIvalInteger ival g
   random g	= randomR (toInteger (minBound::Int), toInteger (maxBound::Int)) g

 An Integer so generated is on the same range as a random Int and uniformly 
 distributed. I could probably exploit this further than I actually do, but 
 using the range feels hacky enough.
 
 It'll be interesting to see what the winning strategy is. Best guess: 
 Construct a nearly sorted list by abusing uniformity to guess index accurately
 and then apply insertion sort to fix any mistakes?
-}

-- Do parallel bucket sort, using constants as above and Data.List.sort in 
-- buckets. Could change precomputed bounds to calls to maximum and minimum...
bucketsort :: [Integer] -> [Integer]
bucketsort = pbucketsort buckets minmax Data.List.sort
  where minmax  = (toInteger (minBound::Int), toInteger (maxBound::Int))
        buckets = 8*1024

-- Test data and benchmark
randomInts = take 200000 (randoms (mkStdGen 211570155)) :: [Integer]
main = defaultMain 
  [ bench "bucketsort"     (nf bucketsort randomInts)
--  , bench "Data.List.sort" (nf sort        randomInts)
  ]
  
-- Parallel bucket sort, with data on some predefined range. 
-- Buckets are sorted with some provided sorting algorithm and concatenated
pbucketsort :: Integer 
           -> (Integer,Integer) 
           -> ([Integer] -> [Integer]) 
           -> [Integer] 
           -> [Integer]
pbucketsort n (xmin,xmax) sort xs = pconcat pdepth bounds $ fmap sort buckets
  where range      = xmax - xmin + 1
        n'         = min n range    -- bound number of buckets
        bucketSize = range `div` n'
        index i    = i `div` bucketSize
        bounds     = ( xmin `div` bucketSize, xmax `div` bucketSize )
        buckets    = bucketize index bounds xs
        pdepth     = floor $ log $ fromIntegral $ 5*numCapabilities

-- Given an indexing function, some range of buckets and input, generate
-- an array of buckets
bucketize :: Ix i => (a -> i) -> (i,i) -> [a] -> Array i [a]
bucketize index bounds xs = 
  accumArray (flip (:)) [] bounds $ map (\x -> (index x,x)) xs
  
-- Parallel concatenation of a subset of an array of lists
-- Try to minimize the expense of concatenation work best we can
pconcat :: (Ix i, Integral i, NFData a) => Int -> (i,i) -> Array i [a] -> [a]
pconcat depth (amin,amax) arr 
  | amin == amax =  arr ! amin
  | depth <= 0   = (arr ! amin) ++ (pconcat depth (amin+1, amax) arr)
  | otherwise    = runEval $ do -- divide, parallelize, concatenate
    let amid = (amax + amin) `div` 2
    ys <- rpar `dot` rdeepseq $ pconcat (depth-1) (amid+1, amax) arr
    xs <- rseq                $ pconcat (depth-1) (amin  , amid) arr
    return $ xs ++ ys
  
  
-- Correctness (behaves like DL.sort)
prop_bucketsortAsSort buckets xs = 
  pbucketsort buckets' minmax sort xs == sort xs
  where types    = ( xs :: [Integer], buckets :: Integer )
        buckets' = (1 + abs buckets) `mod` (2^16)
        minmax   = if null xs then (0,0) else ( minimum xs, maximum xs )